#include <stdlib.h>
#include <stdio.h>
#include <GL/glut.h>
#include <GL/glu.h>
#include <unistd.h>
#include <math.h>
#include "labyrinth.h"
#include "textures.h"
#include "labyrinth_math.h"
#include "maps.h"
#include "key_handler.h"
#include "mouse_handler.h"
#include "fps.h"
#include "labyrinth_timer.h"
#include "draw.h"

ball b = {0.0, 0.0, 0.0, 0.025, (0.025 * 2), (0.025 * 2) * M_PI, 0.0, 0.0, 0.0, 0.0, 0.0}; /* our ball */
boxwall bw = {(BOARD_WIDTH * SQUARE_SIZE) + (2 * SPACE) + SQUARE_SIZE, (SQUARE_SIZE * 5), SQUARE_SIZE}; /* wall of the box */
board bo = {0.0, 0.0}; /* the board */
state s = {0, 0, 0, 0, 0, 0, 0, 0}; /* states */
labyrinth lab = {0, 0.0, 0.0, -4.0, 45.0, 0.0, 0.0, -2, -2}; /* info */

GLfloat light_position[] = {0.0, 0.5, 20.0, 1.0}; /* position of the light */
GLfloat white_light[] = {1.0, 1.0, 1.0, 1.0}; /* color of the light */
GLfloat amb_light[] = {0.3, 0.3, 0.3, 1.0}; /* color of the ambient light */

GLUquadricObj *q; /* quadricobject for the ball and the knobs */

extern int map[BOARD_WIDTH + 1][BOARD_WIDTH + 1];
extern GLuint texture[NUM_TEXTURES];
extern fps lab_fps;

/**
 * Zoom in
 */
void zoomIn()
{
	lab.zoomFactor += 0.01;
}

/**
 * Zoom out
 */
void zoomOut()
{
	lab.zoomFactor -= 0.01;
}

/**
 * The init function
 */
void init(int width, int height)
{
	loadTextures();
	glClearColor(0.9, 0.9, 0.9, 1.0);
	glClearDepth(1.0);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);

	glLightfv(GL_LIGHT0, GL_POSITION,light_position);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, white_light);
	glLightfv(GL_LIGHT0, GL_AMBIENT, amb_light);

	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);

	q = gluNewQuadric();

	gluQuadricNormals(q, GL_SMOOTH);
	gluQuadricTexture(q, GL_TRUE);

	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
}

/**
 * idle function. Takes care of the calculating of the speed, rotation of the board etc.
 */
void idle()
{
	int x,z;
	float framediff;
	lab.currentframe = getsec();
	framediff = lab.currentframe - lab.lastframe;

	if(s.left_p){
		if(bo.rotZ < MAX_ANGLE){
			bo.rotZ += ROT_SPEED * framediff;
		} else{
			bo.rotZ = MAX_ANGLE;
		}
	}

	if(s.up_p){
		if(bo.rotX < MAX_ANGLE){
			bo.rotX += ROT_SPEED * framediff;
		} else{
			bo.rotX = MAX_ANGLE;
		}
	}

	if(s.right_p){
		if(bo.rotZ > -MAX_ANGLE){
			bo.rotZ -= ROT_SPEED * framediff;
		} else{
			bo.rotZ = -MAX_ANGLE;
		}
	}

	if(s.down_p){
		if(bo.rotX > -MAX_ANGLE){
			bo.rotX -= ROT_SPEED * framediff;
		} else{
			bo.rotX = -MAX_ANGLE;
		}
	}

	/* calculate the speed of the ball at this angle */
	b.curSpeedX += -GRAVITY * sin(deg2rad(bo.rotZ)) * framediff;

	/* calculate the speed of the ball at this angle */
	b.curSpeedZ += -GRAVITY * sin(deg2rad(bo.rotX)) * framediff;

	/*
		collision detection
		left side of the board: -0.75
		right side of the board: 0.75
		top of the board: -0.75
		top of the board: 0.75
		if the ball is going left, lets figure out if there is a wall on the left side
		of the ball, and then check the rest of the sides
	*/
	if(b.curSpeedX < 0){ /* going left */
		x = (int)(((b.x - b.rad) / SQUARE_SIZE) + (BOARD_WIDTH / 2));
		z = (int)((b.z / SQUARE_SIZE) + (BOARD_WIDTH / 2));
		if(map[z][x] == 2){
			b.curSpeedX = 0;
			b.x = (x * SQUARE_SIZE) - (SQUARE_SIZE * BOARD_WIDTH / 2) + SQUARE_SIZE + b.rad;
		}
	} else{ /* going right */
		x = (int)(((b.x + b.rad) / SQUARE_SIZE) + (BOARD_WIDTH / 2));
		z = (int)((b.z / SQUARE_SIZE) + (BOARD_WIDTH / 2));
		if(map[z][x] == 2){
			b.curSpeedX = 0;
			b.x = (x * SQUARE_SIZE) - (SQUARE_SIZE * BOARD_WIDTH / 2) - b.rad;
		}
	}

	if(b.curSpeedZ < 0){ /* going up */
		x = (int)((b.x / SQUARE_SIZE) + (BOARD_WIDTH / 2));
		z = (int)(((b.z - b.rad) / SQUARE_SIZE) + (BOARD_WIDTH / 2));
		if(map[z][x] == 2){
			b.curSpeedZ = 0;
			b.z = (z * SQUARE_SIZE) - (SQUARE_SIZE * BOARD_WIDTH / 2) + SQUARE_SIZE + b.rad;
		}
	} else{ /* going down */
		x = (int)((b.x / SQUARE_SIZE) + (BOARD_WIDTH / 2));
		z = (int)(((b.z + b.rad) / SQUARE_SIZE) + (BOARD_WIDTH / 2));
		if(map[z][x] == 2){
			b.curSpeedZ = 0;
			b.z = (z * SQUARE_SIZE) - (SQUARE_SIZE * BOARD_WIDTH / 2) - b.rad;
		}
	}

	b.x += b.curSpeedX;
	b.z += b.curSpeedZ;

	/* rotation */
	b.rotZ = (b.x / b.circumf) * 360;
	b.rotX = (b.z / b.circumf) * 360;

	if(s.zoom_in){
		zoomIn();
	}

	if(s.zoom_out){
		zoomOut();
	}

	lab.lastframe = getsec();
	glutPostRedisplay();
}

/**
 * The draw function. Draws everything on the screen
 */
void draw()
{
	int x, y, z, i;
	GLfloat mat_amb_diff[] = {0.66, 0.39, 0.06, 1.0};
	GLfloat ball[] = {0.3, 0.3, 0.3, 1.0};
	GLfloat knob[] = {0.1, 0.1, 0.1, 1.0};
	GLfloat wall_color[] = {0.1, 0.1, 0.1, 1.0};

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	/* zoom in / out */
	glTranslatef(0.0, 0.0, lab.zoomFactor);

	/* rotate the screen */
	glRotatef(lab.screenRotX, 1.0, 0.0, 0.0);
	glRotatef(lab.screenRotY, 0.0, 1.0, 0.0);
	glRotatef(lab.screenRotZ, 0.0, 0.0, 1.0);

	/* draw the box */
	glPushMatrix();
		/* translate to the top left corner of the frontside of the front wall of the box */
		glTranslatef(-(BOARD_WIDTH * SQUARE_SIZE / 2) - SPACE - bw.depth, 0.0, ((BOARD_WIDTH / 2) * SQUARE_SIZE) + SPACE + bw.depth);
		glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE, mat_amb_diff);
		for(i = 0;i < 4;i++){
			glBindTexture(GL_TEXTURE_2D, texture[i]);
			!s.wireframe ? glBegin(GL_QUADS) : glBegin(GL_LINE_LOOP);
				/* front */
				glNormal3f(0.0, 0.0, 1.0);
				glTexCoord2f(0.0, 1.0); glVertex3f(0.0, 0.0, 0.0);
				glTexCoord2f(1.0, 1.0); glVertex3f(bw.width, 0.0, 0.0);
				glTexCoord2f(1.0, 0.0); glVertex3f(bw.width, -bw.height, 0.0);
				glTexCoord2f(0.0, 0.0); glVertex3f(0.0, -bw.height, 0.0);
			glEnd();
			!s.wireframe ? glBegin(GL_QUADS) : glBegin(GL_LINE_LOOP);
				/* back */
				glNormal3f(0.0, 0.0, -1.0);
				glTexCoord2f(0.0, 1.0); glVertex3f(bw.width, 0.0, -bw.depth);
				glTexCoord2f(1.0, 1.0); glVertex3f(0.0, 0.0,  -bw.depth);
				glTexCoord2f(1.0, 0.0); glVertex3f(0.0, -bw.height,  -bw.depth);
				glTexCoord2f(0.0, 0.0); glVertex3f(bw.width, -bw.height,  -bw.depth);
			glEnd();
			!s.wireframe ? glBegin(GL_QUADS) : glBegin(GL_LINE_LOOP);
				/* top */
				glNormal3f(0.0, 1.0, 0.0);
				glTexCoord2f(0.0, 1.0); glVertex3f(0.0, 0.0, -bw.depth);
				glTexCoord2f(1.0, 1.0); glVertex3f(bw.width, 0.0, -bw.depth);
				glTexCoord2f(1.0, 0.0); glVertex3f(bw.width, 0.0, 0.0);
				glTexCoord2f(0.0, 0.0); glVertex3f(0.0, 0.0, 0.0);
			glEnd();
			!s.wireframe ? glBegin(GL_QUADS) : glBegin(GL_LINE_LOOP);
				/* bottom */
				glNormal3f(0.0, -1.0, 0.0);
				glTexCoord2f(0.0, 1.0); glVertex3f(0.0, -bw.height, 0.0);
				glTexCoord2f(1.0, 1.0); glVertex3f(bw.width, -bw.height, 0.0);
				glTexCoord2f(1.0, 0.0); glVertex3f(bw.width, -bw.height, -bw.depth);
				glTexCoord2f(0.0, 0.0); glVertex3f(0.0, -bw.height, -bw.depth);
			glEnd();
			!s.wireframe ? glBegin(GL_QUADS) : glBegin(GL_LINE_LOOP);
				/* left */
				glNormal3f(-1.0, 0.0, 0.0);
				glTexCoord2f(0.0, 1.0); glVertex3f(0.0, 0.0, -bw.depth);
				glTexCoord2f(1.0, 1.0); glVertex3f(0.0, 0.0, 0.0);
				glTexCoord2f(1.0, 0.0); glVertex3f(0.0, -bw.height, 0.0);
				glTexCoord2f(0.0, 0.0); glVertex3f(0.0, -bw.height, -bw.depth);
			glEnd();
			!s.wireframe ? glBegin(GL_QUADS) : glBegin(GL_LINE_LOOP);
				/* right */
				glNormal3f(1.0, 0.0, 0.0);
				glTexCoord2f(0.0, 1.0); glVertex3f(bw.width, 0.0, 0.0);
				glTexCoord2f(1.0, 1.0); glVertex3f(bw.width, 0.0, -bw.depth);
				glTexCoord2f(1.0, 0.0); glVertex3f(bw.width, -bw.height, -bw.depth);
				glTexCoord2f(0.0, 0.0); glVertex3f(bw.width, -bw.height, 0.0);
			glEnd();
			/* translate to the top left corner of the front part of the next wall */
			glTranslatef(0.0, 0.0, -bw.width - SQUARE_SIZE);
			/* rotate negative 90 degrees around the y axis */
			glRotatef(-90.0, 0.0, 1.0, 0.0);
		}
	glPopMatrix();

	/* draw the knobs using a cylinder and a disc */
	/* knob on the front */
	glDisable(GL_TEXTURE_2D);
	glPushMatrix();
		/* translate to the middle of the frontside of the box */
		glTranslatef(0.0, -bw.height / 2, ((BOARD_WIDTH / 2) * SQUARE_SIZE) + SPACE + bw.depth);
		glRotatef(bo.rotZ, 0.0, 0.0, 1.0);
		glMaterialfv(GL_FRONT,GL_AMBIENT_AND_DIFFUSE, knob);
		gluCylinder(q, SQUARE_SIZE / 2, SQUARE_SIZE / 2, SQUARE_SIZE / 2, 20, 20);
		glTranslatef(0.0, 0.0, SQUARE_SIZE / 2);
		gluDisk(q, 0.0, SQUARE_SIZE / 2, 20, 1);
	glPopMatrix();
	glPushMatrix();
		glTranslatef(bw.width / 2 + bw.depth, -bw.height / 2, 0.0);
		glRotatef(bo.rotX, -1.0, 0.0, 0.0);
		glRotatef(-90, 0.0, 1.0, 0.0);
		glMaterialfv(GL_FRONT,GL_AMBIENT_AND_DIFFUSE, knob);
		gluCylinder(q, SQUARE_SIZE / 2, SQUARE_SIZE / 2, SQUARE_SIZE / 2, 20, 20);
		gluDisk(q, 0.0, SQUARE_SIZE / 2, 20, 1);
	glPopMatrix();
	glEnable(GL_TEXTURE_2D);

	/* rotate so the board and the ball will be drawn in the correct angle */
	/* printf("x: %f\nz: %f\n",bo.rotX,bo.rotZ);*/
	glRotatef(bo.rotX, -1.0, 0.0, 0.0);
	glRotatef(bo.rotZ, 0.0, 0.0, 1.0);

	/* draw the ball */
	glPushMatrix();
		/* translate to the center of the ball */
		glTranslatef(b.x, b.y + b.rad - (SQUARE_SIZE / 2), b.z);
		/* rotate the ball so it appears that the ball is spinning (shows only when there is a texture on the ball) */
		glRotatef(b.rotX, 1.0, 0.0, 0.0);
		glRotatef(b.rotZ, 0.0, 0.0, 1.0);
		/* draw the ball */
		glDisable(GL_TEXTURE_2D);
		glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE, ball);
		gluSphere(q, b.rad, 20, 20);
		glEnable(GL_TEXTURE_2D);
	glPopMatrix();


	glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE, mat_amb_diff);

	glPushMatrix();
		glTranslatef(-BOARD_WIDTH * SQUARE_SIZE / 2, -SQUARE_SIZE / 2, -BOARD_WIDTH * SQUARE_SIZE / 2);
		for(y = 0;y < 2;y++){
			if(y == 0){
				for(z = 0;z < BOARD_WIDTH;z++){
					for(x = 0;x < BOARD_WIDTH;x++){
						if(map[z][x] == 1 || map[z][x] == 2 || map[z][x] == 3){
							glPushMatrix();
								glTranslatef(x * SQUARE_SIZE, y * SQUARE_SIZE, z * SQUARE_SIZE);
								drawFloor(x, z);
							glPopMatrix();
						} else{
							glPushMatrix();
								glTranslatef(x * SQUARE_SIZE, y * SQUARE_SIZE, z * SQUARE_SIZE);
								drawHole();
							glPopMatrix();
						}
					}
				}
			} else{
				for(z = 0;z < BOARD_WIDTH;z++){
					for(x = 0;x < BOARD_WIDTH;x++){
						if(map[z][x] == 2){
							glPushMatrix();
								glTranslatef(x * SQUARE_SIZE, y * SQUARE_SIZE, z * SQUARE_SIZE);
								/* disable texturing to get the correct color on the walls */
								glDisable(GL_TEXTURE_2D);
								glMaterialfv(GL_FRONT,GL_AMBIENT_AND_DIFFUSE, wall_color);
								drawWall();
								/* enable texturing */
								glEnable(GL_TEXTURE_2D);
							glPopMatrix();
						}
					}
				}
			}
		}
	glPopMatrix();

	/* draw some walls around the stage */
	/* first, lets translate ourselves down to one of the corners */
	glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE, mat_amb_diff);
	glTranslatef(-BOARD_WIDTH / 2 * SQUARE_SIZE, -SQUARE_SIZE / 2, BOARD_WIDTH / 2 * SQUARE_SIZE);
	glPushMatrix();
		for(i = 0;i < 4;i++){
			glBindTexture(GL_TEXTURE_2D,texture[4]);
			!s.wireframe ? glBegin(GL_QUADS) : glBegin(GL_LINE_LOOP);
				glNormal3f(0.0, 0.0, 1.0);
				glTexCoord2d(0.0, 1.0); glVertex3f(0.0, 0.0, 0.0);
				glTexCoord2d(1.0, 1.0); glVertex3f(BOARD_WIDTH * SQUARE_SIZE, 0.0, 0.0);
				glTexCoord2d(1.0, 0.0); glVertex3f(BOARD_WIDTH * SQUARE_SIZE, -SQUARE_SIZE, 0.0);
				glTexCoord2d(0.0, 0.0); glVertex3f(0.0, -SQUARE_SIZE, 0.0);
			glEnd();
			glTranslatef(0.0, 0.0, -BOARD_WIDTH * SQUARE_SIZE);
			glRotatef(-90.0, 0.0, 1.0, 0.0);
		}
	glPopMatrix();

	lab_fps.framecount++;

	if(s.verbose){
		lab_fps.curr_time = getsec();
		if((int)lab_fps.curr_time != lab_fps.frame_time){
			printf("%d frames per second.\n", lab_fps.framecount);
			lab_fps.framecount = 0;
			lab_fps.frame_time = lab_fps.curr_time;
		}
	}

	glutSwapBuffers();
}

/**
 * Function that resizes the window
 */
void resize(GLsizei width, GLsizei height)
{
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0,(GLfloat)width/(GLfloat)height,0.1,100.0);
	glMatrixMode(GL_MODELVIEW);
}

/**
 * main function
 */
int main(int argc, char **argv)
{
	int width = 600;
	int height = 600;

	if(argc != 2){
		printf("Please specify a map.\n");
		exit(0);
	}

	if(!loadMap(argv[1])){
		printf("%s could not be loaded\n",argv[1]);
		exit(0);
	}

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_ALPHA | GLUT_DEPTH);
	glutInitWindowSize(width, height);
	glutInitWindowPosition(100, 100);
	lab.window = glutCreateWindow("Labyrinth");
	glutDisplayFunc(draw);
	glutReshapeFunc(resize);
	glutIdleFunc(idle);
	glutKeyboardFunc(keyhandler);
	glutKeyboardUpFunc(keyuphandler);
	glutSpecialFunc(skeyhandler);
	glutSpecialUpFunc(skeyuphandler);
	glutMouseFunc(mouseClick);
	glutMotionFunc(mouseMovement);
	init(width,height);
	glutMainLoop();
	return 0;
}
